process.env['NODE_TLS_REJECT_UNAUTHORIZED'] = '0';

import { readFile } from 'fs/promises';
import got from 'got';
import { JSDOM } from 'jsdom';
import { CookieJar } from 'tough-cookie';
import { promisify }  from 'util';

/**Import conf.json */
const conf = JSON.parse(
  await readFile(
    new URL('./conf.json', import.meta.url)
  )
);

const webhook = conf.webhook;          
const username = conf.username;
const password = conf.password;


const url_cas = "https://cas.ensiie.fr/login"
const url_aurion = "https://aurionweb.ensiie.fr/faces/MainMenuPage.xhtml"



let cookieJar = null;
let aurionSession = null;
let gradesDone = new Map(); //Notes déjà traitées (Map code: note)

/**
 * Request vers aurion / CAS
 * @param {*} url 
 * @param {*} data (Data form, si data alors method->GET)
 * @param {*} method (GET par défaut)
 * @returns Promise<Response>
 */
function aurionRequest(url, data=null, method='GET' ){
    var options = {
        method: method,
	https:{rejectUnauthorized: false},
        cookieJar: cookieJar
    }

    if(data){
        options.form = data;
        options.method = 'POST';
    }

    const promise = got(url, options);
    promise.on('redirect', (updatedOptions, t)=>{
        updatedOptions.method = 'GET';              //302 -> On passe à GET !!
    });

    return promise.then((reponse)=>{
        return new JSDOM(reponse.body).window.document;
    });
}


// Recupere le champs <input> 'execution' de la page de connexion
async function getExecution(){
    const document = await aurionRequest(url_cas);
    return document.querySelector("input[name='execution']").value;
}

//Login avec CAS
async function login(execution){
    let data = {
        'username': username,
        'password': password,
        'execution': execution,
        '_eventId': 'submit',
        'geolocation': ''
    }
    return aurionRequest(url_cas, data=data);
}

//Recupere l'id du menu
function parseMenuID(document){
    let scripts = document.querySelectorAll('script');
    const field = '{PrimeFaces.ab({s:"';
    
    for (const script of scripts){
        if(script.innerHTML && script.innerHTML.includes('chargerSousMenu')){
            let start = script.innerHTML.indexOf(field);
            if(start > 0){
                start += field.length;
                const end = script.innerHTML.indexOf('"', start);
                return script.innerHTML.substring(start, end);
            }
        }
    }
    throw '[PARSE_MENU_ID ERROR] -> Aucun id';
}

//Parse une page et recupere l'id du menu 'Résultats'
function parseMenuResultatID(document){
    const menuItems = document.querySelectorAll('span.ui-menuitem-text');
    for(const menu of menuItems){
        if(menu.innerHTML === 'Résultats'){
            for (const classMenu of menu.parentNode.parentNode.classList){
                if(classMenu.startsWith('submenu_')){
                    return classMenu;
                }
            }
        }
    }
    throw '[PARSE_MENU_RESULTAT_ID ERROR] -> Aucun id';
}

//Recupere l'id du sous menu Résultats > Notes de cette année
function parseSubMenuResultatID(document){
    const updateText = document.querySelector('update').innerHTML.replace('<!--[CDATA[', '').replace(']]', '');
    const newDocument = new JSDOM(updateText).window.document;
    const subMenuItems = newDocument.querySelectorAll('span.ui-menuitem-text');
    const field = "form:sidebar_menuid':'";

    for(const subMenu of subMenuItems){
        if(subMenu.innerHTML === 'Notes aux épreuves année en cours'){
            const script = subMenu.parentNode.getAttribute('onclick');
            const start = script.indexOf(field) + field.length;
            const end = script.indexOf("'", start);
            return script.substring(start, end);
        }
    }
    throw '[PARSE_SUBMENU_RESULTAT_ID ERROR] -> Aucun id';
}

//Récupere le viewState
function parseViewState(document){
    const element = document.querySelector("update[id='j_id1:javax.faces.ViewState:0']");

    if(element){
        return element.innerHTML.replace('<!--[CDATA[', '').replace(']]-->'); //pas sur de ça
    }else{
        return document.querySelector("input[id='j_id1:javax.faces.ViewState:0']").value;
    } 
    
}

//Récupere les sous menus du menu 'résultats'
async function lookForSubResultatMenu(){
    let data = {
        'javax.faces.partial.ajax': 'true',
        'javax.faces.source': aurionSession.formID,
        'javax.faces.partial.execute': aurionSession.formID,
        'javax.faces.partial.render': 'form:sidebar',
        'webscolaapp.Sidebar.ID_SUBMENU':aurionSession.formResultatID,
        'form': 'form',
        'javax.faces.ViewState': aurionSession.viewState
    }
    data[aurionSession.formID] = aurionSession.formID;

    const document = await aurionRequest(url_aurion, data=data);
    aurionSession.formSubResultatID = parseSubMenuResultatID(document);
}

/**
 * Traitement de chaque notes
 * @param {*} gradeElement  (Ligne du tableau correspondant à une note)
 */
async function processGrade(gradeElement){
    const fields = gradeElement.querySelectorAll('td');
    const code = fields[1].innerHTML;
    const title = fields[2].innerHTML;
    const grade = fields[3].innerHTML;


    if(!gradesDone.has(code)){
        gradesDone.set(code, grade);
        await sendDiscordAlert(`${title} : ${grade}`);
    }else if(gradesDone.get(code) !== grade){ //Affiche une mise à jour de note
        await sendDiscordAlert(`[UPDATE] ${title} : ${gradesDone.get(code)} -> ${grade}`);
        gradesDone.set(code, grade);
    }
}

/**
 * Request pour avoir les notes
 */
async function requestGrades(){
    let data = {
        'form': 'form',
        'form:sidebar': 'form:sidebar',
        'form:sidebar_menuid': aurionSession.formSubResultatID,
        'javax.faces.ViewState': aurionSession.viewState
    }

    const document = await aurionRequest(url_aurion, data=data);
    const gradeList = document.querySelector('tbody').querySelectorAll('tr');
    
    for(const grade of gradeList){
        await processGrade(grade);
    }
}
/**
 * Prépare pour récuperer les notes
 */
async function lookForGrades(){
    aurionSession = {};
    let document = await aurionRequest(url_aurion);

    aurionSession.formID = parseMenuID(document);
    aurionSession.formResultatID = parseMenuResultatID(document);
    aurionSession.viewState = parseViewState(document);

    await lookForSubResultatMenu();
    await requestGrades();
}

/**
 * Send alert discord
 * @param {*} message 
 * @returns Promise
 */
function sendDiscordAlert(message){
    let data = {
        'content': message,
        'username': 'Notobot',
    }
    console.log(message);
    return got(webhook, {method:'POST', json:data});
}



async function start(){
    if(!(username && password && webhook)){
        console.log("username, password ou webhook est manquant dans conf.json");
        process.exit(-1);
    }
    while(true){
        try{
            cookieJar = new CookieJar();
            const date = new Date();
            console.log(`[START] \t => \t ${date.toLocaleTimeString('fr')} : ${date.toLocaleDateString('fr')}`);
            
            const execution = await getExecution();
            await login(execution);
            await lookForGrades();
        }catch(e){
            if(e.response && e.response.statusCode === 401){
                console.error('[ERROR 401] -> Bad password / username ?');
            }else{
                console.error(e);
            }
        }finally{
            await new Promise(resolve => setTimeout(resolve,1000*60*15)); //Attente de 15 min
        }
    }

}

start();
